/*
 * $Id: CacheUtils.java 286 2005-12-03 01:07:16Z davidb $
 * 
 * Copyright (c) 2005 VeriSign. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN
 * NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *  
 */

package se.rfc.unbound.cache;

import java.util.Date;
import java.util.Iterator;
import java.util.List;

import org.xbill.DNS.NSRecord;
import org.xbill.DNS.RRSIGRecord;
import org.xbill.DNS.RRset;
import org.xbill.DNS.Record;
import org.xbill.DNS.Section;
import org.xbill.DNS.Type;

import se.rfc.unbound.SMessage;
import se.rfc.unbound.SRRset;

/**
 * This is a collection of utilities that may be used by different cache
 * implementations.
 * 
 * @author davidb
 * @version $Revision: 286 $
 */
public class CacheUtils
{

  /**
   * Given a "glue" rrset, determine if it is relevant to the given NS rrset.
   * I.e., the glue_rrset ownername is one of the NS record targets.
   * 
   * @param ns_rrset The NS rrset to compare against.
   * @param glue_rrset The address rrset to compare.
   * @return true if the glue rrset is relevant.
   */
  public static boolean isRelevantGlue(SRRset ns_rrset, RRset glue_rrset)
  {
    if (ns_rrset.getType() != Type.NS)
      throw new IllegalArgumentException("rrset of non-NS type used.");

    if (glue_rrset.getType() != Type.A && glue_rrset.getType() != Type.AAAA)
    {
      return false;
    }

    for (Iterator i = ns_rrset.rrs(); i.hasNext();)
    {
      NSRecord ns = (NSRecord) i.next();
      if (ns.getTarget().equals(glue_rrset.getName())) return true;
    }
    return false;
  }

  /**
   * Given a List of SRRsets, generate a referral message from it. Generally,
   * this List will come from the readHints() method. The first SRRset must be
   * an NS rrset. Note that if a SRRset in the list is considered relevant
   * (that is, it's name matches the name of a target of one of the NS
   * records, and it is of type A or AAAA), it will be included in the
   * delegation point -- i.e. not discard policy is in place.
   * 
   * @param rrsets The list of SRRsets to process.
   * 
   * @return a SMessage that will be a referral type message.
   */

  public static SMessage referralFromList(List rrsets)
  {
    SMessage m = new SMessage();
    SRRset ns_rrset = null;

    for (Iterator i = rrsets.iterator(); i.hasNext();)
    {
      SRRset rrset = (SRRset) i.next();
      if (rrset.getType() == Type.NS)
      {
        ns_rrset = rrset;
        m.addRRset(ns_rrset, Section.AUTHORITY);
        continue;
      }
      if (rrset.getType() == Type.NSEC || rrset.getType() == Type.DS)
      {
        m.addRRset(rrset, Section.AUTHORITY);
        continue;
      }
      if ((rrset.getType() == Type.A || rrset.getType() == Type.AAAA)
          && isRelevantGlue(ns_rrset, rrset))
      {
        m.addRRset(rrset, Section.ADDITIONAL);
      }
    }

    return m;
  }

  /** Find the minimum actual RRSIG ttl. */
  private static long minSigTTL(RRset rrset)
  {
    long ttl = -1;
    for (Iterator i = rrset.sigs(); i.hasNext(); )
    {
      Record r = (Record) i.next();
      if (ttl < 0 || r.getTTL() < ttl)
      {
        ttl = r.getTTL();
      }
    }
    
    return ttl;
  }
  
  /** Find the maximum signature expiration TTL. */
  private static long maxSigExp(RRset rrset, Date now)
  {
    Date max_exp = null;
    for (Iterator i = rrset.sigs(); i.hasNext(); )
    {
      RRSIGRecord r = (RRSIGRecord) i.next();
      
      if (max_exp == null || r.getExpire().after(max_exp))
      {
        max_exp = r.getExpire();
      }
    }
    
    if (max_exp == null) return -1;
    if (now.after(max_exp)) return 0;
    return (max_exp.getTime() - now.getTime()) / 1000;
  }
  
  /**
   * Given an (S)RRset, trim the TTL of that RRset to the smallest of
   * (rrset.ttl, sigs.ttl, max_sig_expiration - now)
   * 
   * @param rrset
   * @param now The current time -- if null, this will be calculated.
   */
  public static void trimRRsetTTL(RRset rrset, Date now)
  {
    if (now == null) now = new Date();
    
    long orig_ttl = rrset.getTTL();
    long sig_ttl = minSigTTL(rrset);
    long exp_ttl = maxSigExp(rrset, now);
    
    if (sig_ttl < 0) sig_ttl = orig_ttl;
    if (exp_ttl < 0) exp_ttl = orig_ttl;
    
    long ttl = Math.min(orig_ttl, Math.min(sig_ttl, exp_ttl));
    if (ttl < orig_ttl)
    {
      rrset.setTTL(ttl);
    }
  }

  /**
   * Given an original TTL for an cached item, and the expiration time (and a
   * notion of now), return the new TTL for the cache item, reduced by the
   * number of seconds the item has been in the cache. This assumes that the
   * 'expiration' time was calucated by adding the original TTL to the system
   * time at cache insertion.
   * 
   * @param expiration The expiration time of the cache item.
   * @param origTTL The original TTL of the cached item (in seconds).
   * @param now The current system time -- if null, this will be calculated.
   * 
   * @return The new TTL
   */
  public static long cacheTTL(Date expiration, long origTTL, Date now)
  {
    if (now == null) now = new Date();

    long start_time_millis = expiration.getTime() - (origTTL * 1000);
    long time_in_cache_secs = (now.getTime() - start_time_millis) / 1000;
    return origTTL - time_in_cache_secs;
  }
}