/*
 * $Id: RequestMap.java 287 2005-12-06 20:38:27Z davidb $
 *
 * Copyright (c) 2005 VeriSign, Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

package se.rfc.unbound.nsclient;

import java.net.InetSocketAddress;
import java.util.*;
import se.rfc.unbound.*;

import org.apache.log4j.Logger;
import org.xbill.DNS.*;

/**
 * This is a request/response correlation map. This class is used to match
 * responses with outstanding requests. It is fairly stringent, in order to
 * minimize spoofing.
 * 
 * Note that this class is strictly concerned with *outbound* requests and
 * matching responses to them.
 * 
 * This map also tracks request timeouts and wakeups. After a timeout, the
 * timed-out request is removed from the map. After a wakeup, it is not. As a
 * consequence, it is possible for a single DNSEvent instance to be in this
 * map multiple times. As a consequence, it should be noted that the event may
 * change target and request details (like ID) as it gets reused for different
 * specific request sendings.
 * 
 * @author davidb
 * @version $Revision: 287 $
 */
public class RequestMap
{
  // The map of active outbound requests.
  private Map               mMap;
  // The current time (in milliseconds since epoch) of the next
  // expiration/wakeup event.
  private long              mNextExpire;

  private Logger            log             = Logger.getLogger(this
                                                .getClass());

  // Standard timeout of 5 seconds
  private static final long DEFAULT_TIMEOUT = 5000;

  /**
   * This class represents a entry in the request map.
   */
  public static class RequestEntry
  {
    // the actual event in the map.
    public DNSEvent   event;
    // the absolute time after which this event should be "woken up"
    public long       wakeup;
    // the absolute time after which this event should be timed-out.
    public long       expire;
    // Response matching criteria/ per write state.
    public MatchState mstate;

    /**
     * Create a new request entry.
     * 
     * @param event Use this event.
     * @param ms The match state for this event.
     */
    public RequestEntry(DNSEvent event, MatchState ms)
    {
      this.event = event;
      this.mstate = ms;

      setTime(System.currentTimeMillis());
    }

    /**
     * Set the time parameters for this entry based on the current time.
     * 
     * @param now In milliseconds, the value of the current time, or at least
     *          the time that the expiration and wakeup times should be
     *          calculated from.
     */
    private void setTime(long now)
    {
      long timeout = mstate.getTimeoutDuration();
      if (timeout == 0) timeout = DEFAULT_TIMEOUT;
      long wake = mstate.getWakeupDuration();

      expire = now + timeout;
      if (wake > 0) wakeup = now + wake;
    }
  }

  public RequestMap()
  {
    mMap = new HashMap();
    mNextExpire = 0;
  }

  /**
   * Add a request event to this map. This must be followed by a setTime()
   * call on the returned RequestEntry.
   * 
   * @param event The event in question.
   * @param mstate The match state that goes with this event.
   * @return The created RequestEntry object.
   */
  public RequestEntry addRequestEvent(DNSEvent event, MatchState mstate)
  {
    Integer id = new Integer(mstate.getRequestID());

    List l = (List) mMap.get(id);
    if (l == null)
    {
      l = new LinkedList();
      mMap.put(id, l);
    }

    RequestEntry re = new RequestEntry(event, mstate);

    log.trace("adding outstanding request with ID = " + id);
    l.add(re);

    return re;
  }

  /**
   * Given an entry, adjust the next expire time.
   * 
   * @param entry The entry in question.
   */
  private void updateNextExpire(RequestEntry entry)
  {
    long next;
    if (entry.wakeup > 0)
      next = StrictMath.min(entry.wakeup, entry.expire);
    else
      next = entry.expire;

    if (next == 0) return;

    if (mNextExpire == 0)
    {
      mNextExpire = next;
      return;
    }

    mNextExpire = StrictMath.min(mNextExpire, next);
  }

  /**
   * Given a response message, match the response to a registered request. If
   * a match is made, then the matched request is removed from the map (thus
   * deactivating it). It will refuse to match against a "closed" event.
   * 
   * @param m The response message.
   * @param sourceAddress The source of the response message.
   * @return The matching request event, with the response attached, or null
   *         if the response doesn't match a request.
   */
  public RequestEntry findRequestEvent(SMessage m,
      InetSocketAddress sourceAddress)
  {
    Integer id = new Integer(m.getHeader().getID());
    log.trace("findRequestEvent: looking for requests with id: " + id);
    List l = (List) mMap.get(id);

    if (l == null)
    {
      log.trace("no outstanding requests with id = " + id);
      return null;
    }

    for (Iterator i = l.iterator(); i.hasNext();)
    {
      RequestEntry re = (RequestEntry) i.next();
      if (re.event.isClosed())
      {
        log.trace("encountered closed event, removing");
        i.remove();
        continue;
      }

//      log
//          .debug("checking resp against likely candidate: req addr = "
//              + re.mstate.getTarget().getSocketAddress() + " vs "
//              + sourceAddress);
//      log.debug("req ID: " + re.mstate.getRequestID() + " vs "
//          + m.getHeader().getID());
//      log.debug("req q: " + re.mstate.getRequestQuestion() + " vs "
//          + m.getQuestion());

      if (re.mstate.getTarget().getSocketAddress().equals(sourceAddress)
          && responseMatches(m, re.mstate.getRequestQuestion()))
      {
        log.trace("matched response to outstanding request.");
        Response response = new Response(m, Response.OK);
        response.setSource(re.mstate.getTarget());
        re.event.setResponse(response);
        i.remove();
        return re;
      }
    }

    log.trace("no outstanding request that actually matches response");
    return null;
  }

  private boolean responseMatches(SMessage m, Record requestQuestion)
  {
    int rcode = m.getHeader().getRcode();
    // Non-NOERROR/NXDOMAIN responses match only on ID.
    if (rcode != Rcode.NOERROR && rcode != Rcode.NXDOMAIN) return true;
    Record q = m.getQuestion();
    if (q == null)
    {
      return false;
    }
    return q.equals(requestQuestion);
  }

  /**
   * Along will collecting a list of expired or woken-up events, this will
   * update the minimum expiration time.
   * 
   * @return A list of "expired" events -- these are events that have actually
   *         totally expired, or have just been woken up.
   */
  public List expiredEvents()
  {
    List res = new LinkedList();

    // Since the entire list of requests will be traversed, we can reset the
    // next expire time.
    mNextExpire = 0;
    long now = System.currentTimeMillis();

    // For each ID ...
    for (Iterator i = mMap.entrySet().iterator(); i.hasNext();)
    {
      Map.Entry entry = (Map.Entry) i.next();

      List l = (List) entry.getValue();
      // For each event with this ID...
      for (Iterator j = l.iterator(); j.hasNext();)
      {
        RequestEntry re = (RequestEntry) j.next();
        if (re.event.isClosed())
        {
          log.debug("removing closed event.");
          j.remove();
          continue;
        }

        // if wakeup!
        if (re.wakeup > 0 && re.wakeup < now)
        {
          re.wakeup = 0; // clear wakeup timer.
          re.event.setResponse(Response.wakeupResponse());
          res.add(re);
        }

        // if expired!
        if (re.expire > 0 && re.expire < now)
        {
          re.expire = 0;
          re.event.setResponse(Response.timeoutResponse());
          res.add(re);
          j.remove();
          continue;
        }

        updateNextExpire(re);
      }

      if (l.size() == 0) i.remove();
    }
    return res;
  }

  /**
   * Calculate and return the next timeout duration, in milliseconds. This is
   * based on the difference between the next expiration/wakeup time and the
   * value of now.
   * 
   * @param now In milliseconds, the current time.
   * @return 0 if there are no outstanding requests (infinite timeout), a
   *         postive duration of the number of milliseconds to delay before
   *         the next timing event, or a negative number, indicating no
   *         timeout.
   */
  public long getNextTimeout(long now)
  {
    // No NextExpire time means no active outbound requests, so the timeout
    // should be 0, for infinite.
    if (mNextExpire == 0) return 0;

    return mNextExpire - now;
  }

}
